#include "moglclass.h"
#include "shaders.h"
#include "geometryPositionNormalTexture.h"
#include "light.h"
#include "material.h"
#include "texture2D.h"
#include "camera.h"

using namespace moglclass;

#define NUMSHADERS 4

Shader shShaders[NUMSHADERS];
Program spMain;

float g_fovy = 45.0f;
float g_aspect = 1.0f;
float g_zNear = 1.0f;
float g_zFar = 100.f;

Camera camera1(FLY);
//float cameraHeading = 0.0f;
//float cameraZ = 10.0f;
glm::vec3 g_eye(0, 0, 10);
glm::vec3 g_at(0, 0, 0);
glm::vec3 g_up(0, 1, 0);

glm::mat4 Identity(1.0f);
glm::mat4 World(1.0f);
glm::mat4 View(1.0f);
glm::mat4 Projection(1.0f);
glm::mat4 MVP(1.0f);
glm::mat4 xyRotation(1.0f);

Light light0(glm::vec4(0.0f, 0.0f, 5.0f, 0.0f), glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), 1.0f, 0.0f, 0.0f, 180.0f, 0.0f, glm::vec3(0.0f, 0.0f, 0.0f)); 
Light light1(glm::vec4(-1.0f, 1.0f, 0.0f, 0.0f), glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), 1.0f, 0.0f, 0.0f, 180.0f, 0.0f, glm::vec3(0.0f, 0.0f, 0.0f));
Light light2(glm::vec4(-5.0f, -5.0f, 4.0f, 0.0f), glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), 0.0f, 1.0f, 0.0f, 60.0f, 50.0f, glm::vec3(0.0f, 0.0f, -1.0f));

Material whiteMaterial(glm::vec4(0.2f, 0.2f, 0.2f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), glm::vec4(1.0f, 1.0f, 1.0f, 1.0f), 50.0f);
float g_alpha = 1.0f;

Texture2D * groundTex = NULL;
Texture2D * grassTex = NULL;
Texture2D * treeTex[4];

Quad * tree = NULL;
Grid * ground = NULL;
Cylinder * cyl = NULL;

bool g_wiremode = false;
bool g_animmode = false;
int startTime = 0;
bool g_fademode = false;
glm::vec3 g_Position1(-3, 0, 1);
glm::vec3 g_Position2(-1, 0, 0.2);
glm::vec3 g_Position3(1, 0, 0.5);
glm::vec3 g_Position4(3, 0, 0);
bool g_billboard = false;
glm::mat4 Billboard(1.0f);

// create a axis-aligned billboard matrix
void buildAxisAlignedBillboardMatrix(glm::mat4& m, glm::vec3& axis)
{
	float pi = 3.141592654f;
	float theta = -atan2f(m[0][2], m[2][2]);
	float ct = cosf(theta);
	float st = sinf(theta);

	// normalize
	axis = glm::normalize(axis);

	// clear out the view matrix passed in
	m = glm::mat4(1.0f);

	//---------------------------------------------
	// R = uu' + cos(theta)*(I-uu') + sin(theta)*S
	//
	// S =  0  -z   y    u' = (x, y, z)
	//	    z   0  -x
	//	   -y   x   0
	//---------------------------------------------
	m[0][0] = axis[0] * axis[0] + ct*(1 - axis[0] * axis[0]) ;
	m[0][1] = axis[0] * axis[1] + ct*(0 - axis[0] * axis[1]) + st*(-axis[2]);
	m[0][2] = axis[0] * axis[2] + ct*(0 - axis[0] * axis[2]) + st*axis[1];

	m[1][0] = axis[1] * axis[0] + ct*(0 - axis[1] * axis[0]) + st*axis[2];
	m[1][1] = axis[1] * axis[1] + ct*(1 - axis[1] * axis[1]) ;
	m[1][2] = axis[1] * axis[2] + ct*(0 - axis[1] * axis[2]) + st*(-axis[0]);

	m[2][0] = axis[2] * axis[0] + ct*(0 - axis[2] * axis[0]) + st*(-axis[1]);
	m[2][1] = axis[2] * axis[1] + ct*(0 - axis[2] * axis[1]) + st*axis[0];
	m[2][2] = axis[2] * axis[2] + ct*(1 - axis[2] * axis[2]) ;
}


void init( void )
{
    // load shaders and use the resulting shader program
	spMain.createProgram();
	shShaders[0].loadShader("MTexturePerPixelShading.vs", GL_VERTEX_SHADER);
	shShaders[1].loadShader("SimpleTextureShading.fs", GL_FRAGMENT_SHADER);
	spMain.attachShaderToProgram(&shShaders[0]);
	spMain.attachShaderToProgram(&shShaders[1]);
	spMain.linkProgram();
	spMain.useProgram();
	
	Projection = glm::perspective(g_fovy, g_aspect, g_zNear, g_zFar);
	View = camera1.lookAt(g_eye, g_at, g_up);

	// texture
	groundTex = new Texture2D("grass-texture.jpg");
	grassTex = new Texture2D("grass.tga");
	treeTex[0] = new Texture2D("tree01S.png");
	treeTex[1] = new Texture2D("tree02S.png");
	treeTex[2] = new Texture2D("tree35S.png");
	treeTex[3] = new Texture2D("tree.tga");

	ground = new Grid(glm::vec3(0.0f, -2.0f, 0.0f), glm::vec3(0.0f, 1.0f, 0.0f), glm::vec3(0.0f, 0.0f, 1.0f), 20.0f, 20.0f, 10, 10); // texture grid
	tree = new Quad(glm::vec3(0.0f, 0.0f, 0.0f), glm::vec3(0.0f, 0.0f, 1.0f), glm::vec3(0.0f, 1.0f, 0.0f), 2.0f, 3.0f); // texture quad
	cyl = new Cylinder(glm::vec3(-1.0, 0, -0.5), 1.0f, 2.0f, 16);

	glEnable(GL_CULL_FACE); // back-face culling
	glEnable(GL_DEPTH_TEST); // depth test
	glDepthFunc(GL_LESS);
	glEnable(GL_ALPHA_TEST); // alpha test
	glAlphaFunc(GL_GREATER, 0.5);

	glClearColor( 1.0, 1.0, 1.0, 1.0 ); // white background
}

void cleanup()
{
	if (tree) delete tree;
	if (ground) delete ground;
	if (cyl) delete cyl;

	spMain.deleteProgram();
	printf("clean up\n");
}

void drawTextureGround(glm::mat4& worldMatrix, glm::vec3& position)
{
	//World = glm::translate(glm::mat4(1.0f), position);
	World = glm::translate(worldMatrix, position);
	spMain.setUniform("gModel", World);
	spMain.setUniform("gTextureSampler", 0);
	groundTex->bind();
	ground->draw();
	groundTex->disable();
}

void drawTextureTree(glm::mat4& worldMatrix, glm::vec3& position, int index)
{
	//World = glm::translate(glm::mat4(1.0f), position);
	World = glm::translate(worldMatrix, position);
	spMain.setUniform("gModel", World);
	spMain.setUniform("gTextureSampler", 0);
	treeTex[index]->bind();
	tree->draw();
	treeTex[index]->disable();
}

void drawTextureGrass(glm::mat4& worldMatrix, glm::vec3& position)
{
	//World = glm::translate(glm::mat4(1.0f), position);
	World = glm::translate(worldMatrix, position);
	spMain.setUniform("gModel", World);
	spMain.setUniform("gTextureSampler", 0);
	grassTex->bind();
	tree->draw();
	grassTex->disable();
}

void display( void )
{
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );     // clear the window

	// set drawmode
	//if (g_wiremode) 
		//glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);  // draw wireframe 
	//else
		//glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);	// draw filled 

	// MVP matrix
	View = camera1.View();
	//World = glm::rotate(glm::mat4(1.0f), -cameraHeading, glm::vec3(0, 1, 0)) * glm::translate(glm::mat4(1.0f), glm::vec3(0, 0, -cameraZ));

	MVP = Projection * View * World;
	spMain.useProgram();
	spMain.setUniform("gMVP", MVP);
	spMain.setUniform("gProjection", Projection);
	spMain.setUniform("gView", View);
	spMain.setUniform("gModel", World);

	glm::vec3 cameraPos = camera1.getPosition();
	//printf("cameraPos: %f %f %f\n", cameraPos[0], cameraPos[1], cameraPos[2]);
	spMain.setUniform("gEyePosition", cameraPos);

	// light
	light0.setUniform(&spMain, "gLight0");
	//light1.setUniform(&spMain, "gLight1");
	//light2.setUniform(&spMain, "gLight2");

	// material
	whiteMaterial.setDiffuse(glm::vec4(1, 1, 1, 1));
	whiteMaterial.setUniform(&spMain, "gMaterial");
	// draw ground
	drawTextureGround(xyRotation, glm::vec3(0, 0, 1));

	// draw trees
	if (g_billboard) {
		Billboard = View;
		// billboard rotation about the y axis
		//glDepthMask(false);
		buildAxisAlignedBillboardMatrix(Billboard, glm::vec3(0.0f, 1.0f, 0.0f));
		drawTextureTree(Billboard, g_Position1, 0);
		drawTextureTree(Billboard, g_Position2, 1);
		drawTextureTree(Billboard, g_Position4, 2);
		drawTextureTree(Billboard, g_Position3, 3);
		//glDepthMask(true);
	}
	else {
		//drawTextureGrass(xyRotation, glm::vec3(0, 0, 0.4));
		drawTextureTree(xyRotation, g_Position1, 0);
		drawTextureTree(xyRotation, g_Position2, 1);
		drawTextureTree(xyRotation, g_Position4, 2);
		drawTextureTree(xyRotation, g_Position3, 3);
	}

	glutSwapBuffers();
}

void update()
{
	static int currentTime, deltaTime, prevTime = 0; 
	static int elapsedTime = 0;
	
	currentTime = glutGet(GLUT_ELAPSED_TIME);
	deltaTime = currentTime - prevTime;
	prevTime = currentTime;

	if (g_fademode)
	{
		g_alpha = 1.0f - fabs(sinf((float)elapsedTime * 0.0002f));
		printf("g_alpha=%f\n", g_alpha);
	}

	if (g_animmode)
	{
		// translate & rotate around x/y axis
		static float angle = 0.0f;
		angle += 0.001f;

		// geometry: translate & rotate in x & rotate in y
		glm::mat4 T = glm::translate(glm::mat4(1.0f), glm::vec3(0, 0, 1));
		glm::mat4 R = glm::rotate(T, angle, glm::vec3(1, 0, 0));
		xyRotation = glm::rotate(R, angle, glm::vec3(0, 1, 0));
	}

	elapsedTime = currentTime - startTime;
	glutPostRedisplay();
}

void keyboard( unsigned char key, int x, int y )
{
	switch (key)  
	{    
	case 0x1B:    
	case 'q':    
	case 'Q':      
		exit(0);     
	case 'b':
		g_billboard = !g_billboard;
		printf("g_billboard=%d\n", g_billboard);
		break;
	case 'm':
		g_animmode = !g_animmode;
		break;  
	case ' ':
		g_fademode = !g_fademode;
		startTime = glutGet(GLUT_ELAPSED_TIME);
		break;  
	}
	glutPostRedisplay();
}
/*
void specialkey(int key, int x, int y )
{
	switch (key)  
	{    
	case GLUT_KEY_LEFT:
		cameraHeading -= 10;
		printf("cameraHeading=%f\n", cameraHeading);
		break;  
	case GLUT_KEY_RIGHT:    
		cameraHeading += 10;
		printf("cameraHeading=%f\n", cameraHeading);
		break;  
	case GLUT_KEY_UP:      
		cameraZ += 1.0f;
		printf("cameraZ=%f\n", cameraZ);
		break;  
	case GLUT_KEY_DOWN:
		cameraZ -= 1.0f;
		printf("cameraZ=%f\n", cameraZ);
		break;  
	case GLUT_KEY_PAGE_UP:      
		break;  
	case GLUT_KEY_PAGE_DOWN:
		break;  
	}
	glutPostRedisplay();
}
*/

void specialkey(int key, int x, int y)
{
 if (key == GLUT_KEY_F1)   // x-movement
        camera1.strafe(0.5);
 else if (key == GLUT_KEY_F2)
        camera1.strafe(-0.5);
 else if (key == GLUT_KEY_F3) // y-movement
        camera1.fly(0.5);
 else if (key == GLUT_KEY_F4)
        camera1.fly(-0.5);
 else if (key == GLUT_KEY_F5) // z-movement
        camera1.walk(0.5);
    else if (key == GLUT_KEY_F6)
        camera1.walk(-0.5);

 else if (key == GLUT_KEY_F7) // yaw (by y-axis)
        camera1.yaw(2.5);
 else if (key == GLUT_KEY_F8)
        camera1.yaw(-2.5);
 else if (key == GLUT_KEY_F9) // pitch (by x-axis)
        camera1.pitch(2.5);
    else if (key == GLUT_KEY_F10)
        camera1.pitch(-2.5);
 else if (key == GLUT_KEY_HOME) // roll (by z-axis)
        camera1.roll(2.5);
    else if (key == GLUT_KEY_END)
        camera1.roll(-2.5);

 else if (key == GLUT_KEY_LEFT) // same as town
        camera1.yaw(2.5);
 else if (key == GLUT_KEY_RIGHT)
        camera1.yaw(-2.5);
 else if (key == GLUT_KEY_UP) 
        camera1.walk(0.5);
    else if (key == GLUT_KEY_DOWN)
        camera1.walk(-0.5);

 glutPostRedisplay();
}


void mouse(int button, int state, int mx, int my)
{
	if (button==GLUT_RIGHT_BUTTON && state==GLUT_DOWN)
		exit(0);
}

void motion(int mx, int my)
{
	int w = glutGet(GLUT_WINDOW_WIDTH);
	int h = glutGet(GLUT_WINDOW_HEIGHT);
	float x = (float) (mx - w*0.5) / w;
	float y = (float) (h*0.5 - my) / h;
	//fprintf(stderr, "x=%f y=%f\n", x, y);

	glutPostRedisplay();
}

void reshape(int w, int h)
{
	g_aspect = (float) (w/h);

    glViewport(0, 0, (GLsizei) w, (GLsizei) h);
    glutPostRedisplay();
}

int main( int argc, char **argv )
{
	glutInit(&argc, argv);
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
    glutInitWindowSize( SCREEN_WIDTH, SCREEN_HEIGHT );
    glutCreateWindow(argv[0]);
    
	glewInit();
    init();

    glutDisplayFunc( display );
    glutKeyboardFunc( keyboard );
    glutSpecialFunc( specialkey );
	glutMouseFunc( mouse ); 
	glutMotionFunc( motion ); 
	glutIdleFunc( update );
	glutReshapeFunc( reshape );

    glutMainLoop();

	cleanup();

    return 0;
}
