#include "multitexQuad.h"

namespace moglclass
{
	MultitexQuad::MultitexQuad(glm::vec3 p_, glm::vec3 n_, glm::vec3 u_, float width_, float height_) : GeometryPositionNormalTexture()
	{
		set(p_, n_, u_, width_, height_);
	}

	void MultitexQuad::set(glm::vec3 p_, glm::vec3 n_, glm::vec3 u_, float width_, float height_)
	{
		p = p_; // center position
		n = n_; // normal vector
		u = u_; // up vector
		width = width_;
		height = height_;
		init();
	}

	void MultitexQuad::init()
	{
		glm::vec3 left = glm::cross(n, u);
		glm::vec3 uppercenter = (u * height / 2.0f) + p;
		glm::vec3 tl = uppercenter + (left * width / 2.0f);
		glm::vec3 tr = uppercenter - (left * width / 2.0f);
		glm::vec3 bl = tl - (u * height);
		glm::vec3 br = tr - (u * height);

		// face 1
		vbo.addData(&bl[0], sizeof(glm::vec3)); // vertex position
		vbo.addData(&n[0], sizeof(glm::vec3)); // vertex normal
		vbo.addData(&glm::vec2(0.0f, 0.0f), sizeof(glm::vec2)); // vertex texture coordnate
		vbo.addData(&glm::vec2(0.0f, 0.0f), sizeof(glm::vec2)); // vertex texture coordnate

		vbo.addData(&br[0], sizeof(glm::vec3));
		vbo.addData(&n[0], sizeof(glm::vec3));
		vbo.addData(&glm::vec2(1.0f, 0.0f), sizeof(glm::vec2));
		vbo.addData(&glm::vec2(1.0f, 0.0f), sizeof(glm::vec2));

		vbo.addData(&tr[0], sizeof(glm::vec3));
		vbo.addData(&n[0], sizeof(glm::vec3));
		vbo.addData(&glm::vec2(1.0f, 1.0f), sizeof(glm::vec2));
		vbo.addData(&glm::vec2(1.0f, 1.0f), sizeof(glm::vec2));

		// face 2
		vbo.addData(&bl[0], sizeof(glm::vec3));
		vbo.addData(&n[0], sizeof(glm::vec3));
		vbo.addData(&glm::vec2(0.0f, 0.0f), sizeof(glm::vec2));
		vbo.addData(&glm::vec2(0.0f, 0.0f), sizeof(glm::vec2));

		vbo.addData(&tr[0], sizeof(glm::vec3));
		vbo.addData(&n[0], sizeof(glm::vec3));
		vbo.addData(&glm::vec2(1.0f, 1.0f), sizeof(glm::vec2));
		vbo.addData(&glm::vec2(1.0f, 1.0f), sizeof(glm::vec2));

		vbo.addData(&tl[0], sizeof(glm::vec3));
		vbo.addData(&n[0], sizeof(glm::vec3));
		vbo.addData(&glm::vec2(0.0f, 1.0f), sizeof(glm::vec2));
		vbo.addData(&glm::vec2(0.0f, 1.0f), sizeof(glm::vec2));

		numVertices = 6;
		createVAO();
		isLoaded = true;
	}

	void MultitexQuad::createVAO()
	{
		// VAO & VBOs
		vbo.createVBO();
		vbo.bindVBO();
		vbo.uploadDataToGPU(GL_STATIC_DRAW);

		glGenVertexArrays(1, &vao);
		glBindVertexArray(vao);

		int iDataStride = 2 * sizeof(glm::vec3) + 2 * sizeof(glm::vec2); // vertex & normal & texCoord0 & texCoord1

		int iDataOffset = 0;
		glEnableVertexAttribArray(0); // position
		glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, iDataStride, (void*)iDataOffset);
		iDataOffset += sizeof(glm::vec3);

		glEnableVertexAttribArray(1); // normal
		glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, iDataStride, (void*)iDataOffset);
		iDataOffset += sizeof(glm::vec3);
		
		glEnableVertexAttribArray(2); // texCoord0
		glVertexAttribPointer(2, 2, GL_FLOAT, GL_FALSE, iDataStride, (void*)iDataOffset);
		iDataOffset += sizeof(glm::vec2);

		glEnableVertexAttribArray(3); // texCoord1
		glVertexAttribPointer(3, 2, GL_FLOAT, GL_FALSE, iDataStride, (void*)iDataOffset);
	}

	void MultitexQuad::draw()
	{
		if (!isLoaded) return;

		glBindVertexArray(vao);
		glDrawArrays(GL_TRIANGLES, 0, numVertices);
	}

}